import Stemmer
import mailbox
import re
import os
import operator
import pickle
import sys
import getopt

dump = 'dump'
stemmer = None

def usage():
    print("Usage: %s options\n" % sys.argv[0])
    print(" -h --help           Display this usage information.\n"
          " -s --stemmer        Use stemmer.\n"
          " -d --dump <path>    Path to saved dictionaries (default: dump).\n")

def parse_opts():
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'hsd:',
                                   ['help','stemmer', 'dump'])
    except getopt.GetoptError, err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        usage()
        sys.exit(1)

    for o, a in opts:
        if o in ("-d", "--dump"):
            global dump
            dump = a
        elif o in ("-s", "--stemmer"):
            global stemmer
            stemmer = Stemmer.Stemmer('english')
        elif o in ("-h", "--help"):
            usage()
            sys.exit()
        else:
            assert False, "unhandled option"

def add_word(hash, w):
    # TODO add some word filtering?
    if hash.has_key(w):
        hash[w] = hash[w] + 1
    else:
        hash[w] = 1

def read_mail(path):
    try:
        f = open(path, 'r')
    except:
        return False
    m = mailbox.Message(f.read())

    msg = None
    # use walk to find parts hidden in one or more encapsulations
    for part in m.walk():
        if part.get_content_type() == 'text/plain':
            msg = part.get_payload()
            break

    # no plain message -- we don't handle html
    if not msg:
        return None

    # add header values -- huge boost
    for val in m.values():
        msg += "\n" + val

    msg = msg.lower()
    words = re.sub('[*+~:.,;\'"<>/\\?!-=@#&_{}()\[\]0-9]', ' ', msg).split()
    return words

def add_words(hash, words):
    if stemmer:
        words = stemmer.stemWords(words)
    for w in words:
        add_word(hash, w)

def add_mail(hash, path):
    words = read_mail(path)
    if words:
        add_words(hash, words)
        return True
    else:
        return False

# calculate spam probability for word
def calculate(ham, spam, word):
    g = 2 * ham[word]
    b = spam[word]
    if g + b >= 5:
        return max(.01, min(.99,
                            min(1.0, float(b)/nbad) / (min(1.0,
                                                           float(g)/ngood) +
                                                       min(1.0,
                                                           float(b)/nbad))))

class Words(dict):
    def __missing__(self, key):
        return 0

# count of mails added as spam or ham
ham = Words()
nbad = 0
spam = Words()
ngood = 0
# calculated probabilities
prob = dict()

def add_data(path, list):
    for f in os.listdir(path):
        list.append(path + '/' + f)

# training+testing data
data_spam = []
add_data('data/spam', data_spam)
data_ham = []
add_data('data/POWERTOP', data_ham)

# prepare test data
data_test_ham = []
add_data('data/ThinkPad', data_test_ham)
# split spam (training and testing set)
c = len(data_spam)/2
data_spam, data_test_spam = data_spam[:c], data_spam[c:]

parse_opts()

# load or build dictionaries
if os.path.exists(dump):
    f = open(dump, 'r')
    ham,spam,prob,nbad,ngood = pickle.load(f)
else:
    # add spam
    for path in data_spam:
        if add_mail(spam, path):
            nbad += 1
    
    # add ham
    for path in data_ham:
        if add_mail(ham, path):
            ngood += 1

    # calculate spam probability for words in both dictionaries
    for word in set(ham.keys() + spam.keys()):
        value = calculate(ham, spam, word)
        if value:
            prob[word] = value

    # save trained dictionaries
    pickle.dump([ham,spam,prob,nbad,ngood], open(dump,'w'))

print 'nbad, ngood: %d, %d' % (nbad, ngood)

# normalize values so we can sort by lowest distance from 0.0 or 1.0
def normalize(key):
    if key[1] <= 0.5:
        return key[1]
    return abs(key[1] - 1)

# calculate probability for mail
def bayes(prob, path, count=15):
    words = read_mail(path)

    # empty mail - skip
    if not words:
        print path
        return None

    # build word list for known words
    sel = []
    for w in set(words):
        if prob.has_key(w):
            sel.append([w, prob[w]])

    # no words - skip
    if len(sel) == 0:
        print path
        return None

    # select N most interesting words (farthest from 0.5)
    sel = sorted(sel, key=normalize)[:count]
    # keep only probabilities
    # TODO we don't need them at all, only for "debugging"
    sel = map(lambda x: x[1], sel)

    # calculate (w1 * w2 * ... * wN) / [(w1 * w2 * ... * wN) + ((1-w1) * ...)]
    prod = reduce(operator.mul, sel)
    res = prod / (prod + reduce(operator.mul, map(lambda x: 1-x, sel)))
    return res

boundary = 0.9
# TEST 1: spam classification
rspam = []
rspamfalse = []
spamfail = 0
for i in data_test_spam:
    res = bayes(prob, i)
    if not res:
        spamfail += 1
        continue
    rspam.append(res)
    if res <= boundary:
        rspamfalse.append([res, i])

# TEST 2: false positives
rham = []
rhamfalse = []
hamfail = 0
for i in data_test_ham:
    res = bayes(prob, i)
    if not res:
        hamfail += 1
        continue
    rham.append(res)
    if res > boundary:
        rhamfalse.append([res, i])

# RESULTS
# spam fail
#print rspamfalse
print len(rspamfalse)
print spamfail
print len(data_test_spam)
# false positives
#print rhamfalse
print len(rhamfalse)
print hamfail
print len(data_test_ham)
